/*
  +-------------------------------------+
  |                                     |
  |   Code to handle Windows sockets.   |
  |                                     |
  +-------------------------------------+
*/
#include <windows.h>
#include <stdio.h>
#include <winsock.h>
#include "print.h"
#include "socks.h"

/*
  +------------------------------------------+
  |                                          |
  |   This class handles initialization of   |
  |   Windows sockets and guarantees that    |
  |   the WSACleanup function gets called    |
  |   when we exit.                          |
  |                                          |
  +------------------------------------------+
*/
wsockets_t::wsockets_t()
{
    dprintf("Initializing WinSock\n");
    ok = 0;
    WORD vers = MAKEWORD( 1, 1);
    WSADATA wsadata;
    int ret = WSAStartup( vers, &wsadata);
    if( ret != 0 ) {
        tprintf("Error: couldn't initialize winsock\n");
        throw WINSOCK_FAILURE;
    }
    if( wsadata.wVersion != vers ) {
        tprintf("Error: winsock version %d.%d not supported",
          (int)LOBYTE( vers), (int)HIBYTE( vers));
        WSACleanup();
        throw WINSOCK_FAILURE;
    }
    ok = 1;
}

wsockets_t::~wsockets_t()
{
    if( ok ) WSACleanup();
}


/*
  +-----------------------------------------------------------+
  |                                                           |
  |   Socket class.  Automagically closes socket              |
  |   when it goes out of scope.  Hides some ugly details.    |
  |   I've simplified this somewhat in that I've combined     |
  |   bind() and listen() since I always want it to do a      |
  |   "bind" followed by a "listen" on the local host, etc.   |
  |   I've also removed parameters that I'm always going to   |
  |   set the same way--maybe later I will have to            |
  |   extend...                                               |
  |                                                           |
  +-----------------------------------------------------------+
*/
socket_t::socket_t()
{
    sd = socket( AF_INET, SOCK_STREAM, 0);
    if( sd == INVALID_SOCKET ) throw BAD_SOCKET;
}

socket_t::~socket_t()
{
    if( sd != INVALID_SOCKET ) closesocket( sd);
}

void socket_t::create()
{
    if( sd != INVALID_SOCKET ) closesocket( sd);
    sd = socket( AF_INET, SOCK_STREAM, 0);
    if( sd == INVALID_SOCKET ) throw BAD_SOCKET;
}

//
// Probably ought to have a copy constructor as well...
//
socket_t &socket_t::operator=( socket_t &s)
{
    if( this == &s ) return *this;
    if( sd != INVALID_SOCKET ) closesocket( sd);
    sd = s.sd;
    s.sd = INVALID_SOCKET;
    return *this;
}

socket_t &socket_t::operator=( SOCKET as)
{
    if( sd != INVALID_SOCKET ) closesocket( sd);
    sd = as;
    // This could let a bad socket in...maybe it should be tested?
    return *this;
}

void socket_t::listen( unsigned short port, int maxqlen)
{
    sockaddr_in srvr_addr;
    srvr_addr.sin_family = AF_INET;
    srvr_addr.sin_addr.s_addr = htonl( INADDR_ANY);
    srvr_addr.sin_port = htons( port);
    int retval = bind( sd, (sockaddr *)&srvr_addr, sizeof( srvr_addr));
    if( retval == SOCKET_ERROR ) throw BIND_FAILURE;
    retval = ::listen( sd, maxqlen);
    if( retval == SOCKET_ERROR ) throw LISTEN_FAILURE;
}

void socket_t::connect( char *host, unsigned short port)
{
    sockaddr_in srvr_addr;
    srvr_addr.sin_family = AF_INET;
    hostent *hstnt = gethostbyname( host);
    if( hstnt == NULL ) throw HOSTNAME_LOOKUP_FAILURE;
    srvr_addr.sin_addr = *(in_addr *)hstnt->h_addr_list[0];
    srvr_addr.sin_port = htons( port);
    int retval = ::connect( sd, (sockaddr *)&srvr_addr, sizeof( srvr_addr));
    if( retval == SOCKET_ERROR ) throw CONNECT_FAILURE;
}

SOCKET socket_t::accept()
{
    SOCKET as = ::accept( sd, NULL, NULL);
    if( as == INVALID_SOCKET ) throw ACCEPT_FAILURE;
    return as;
}

void socket_t::close()
{
    closesocket( sd);
    sd = INVALID_SOCKET;
}

void socket_t::getpeername( char *name, int nmsz)
{
    sockaddr_in srvr_addr;
    srvr_addr.sin_family = AF_INET;
    int len = sizeof( sockaddr_in);
    int retval = ::getpeername( sd, (sockaddr *)&srvr_addr, &len);
    if( retval != 0 ) throw GETPEERNAME_FAILURE;
    strncpy( name, inet_ntoa( srvr_addr.sin_addr), nmsz);
}

/*
  +-----------------------------+
  |                             |
  |   Miscelaneous functions.   |
  |                             |
  +-----------------------------+
*/

//
// Return the IP address of the host given by hostname.
//  (Should I worry about the newer INET6 addresses???)
//
int get_ip_address( char *hostname, in_addr *inaddr)
{
    hostent *hstnt = gethostbyname( hostname);
    if( hstnt == NULL ) return 0;
    *inaddr = *(in_addr *)hstnt->h_addr_list[0];
    return 1;
}

//
// Print out local host name and address.
//
void print_localhost_info()
{
    char thishost[256];
    if( gethostname( thishost, 256) != 0 ) {
        tprintf("Error: couldn't get local hostname\n");
        throw HOSTNAME_LOOKUP_FAILURE;
    }
    dprintf("Local hostname: %s\n", thishost);
    in_addr inaddr;
    if( !get_ip_address( thishost, &inaddr) ) {
        tprintf("Error: couldn't get local IP address\n");
        throw IPADDRESS_LOOKUP_FAILURE;
    }
    dprintf("Local IP address: %s\n", inet_ntoa( inaddr));
}
